"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var Type_1 = require("./Type");
var Support_1 = require("./Support");
var REQUIRED_OVERLAP = 3 / 4;
function canBeCombined(c1, c2) {
    var p1 = c1.properties;
    var p2 = c2.properties;
    if (p1.size < p2.size * REQUIRED_OVERLAP || p2.size < p1.size * REQUIRED_OVERLAP) {
        return false;
    }
    var larger;
    var smaller;
    if (p1.size > p2.size) {
        larger = p1;
        smaller = p2;
    }
    else {
        larger = p2;
        smaller = p1;
    }
    var minOverlap = Math.ceil(larger.size * REQUIRED_OVERLAP);
    var maxFaults = smaller.size - minOverlap;
    Support_1.assert(maxFaults >= 0, "Max faults negative");
    var commonProperties = [];
    var faults = 0;
    smaller.forEach(function (_, name) {
        if (larger.has(name)) {
            commonProperties.push(name);
        }
        else {
            faults += 1;
            if (faults > maxFaults)
                return false;
        }
    });
    if (faults > maxFaults)
        return false;
    for (var _i = 0, commonProperties_1 = commonProperties; _i < commonProperties_1.length; _i++) {
        var name = commonProperties_1[_i];
        var ts = smaller.get(name);
        var tl = larger.get(name);
        if (ts === undefined || tl === undefined) {
            return Support_1.panic("Both of these should have this property");
        }
        var tsCases = Type_1.nonNullTypeCases(ts);
        var tlCases = Type_1.nonNullTypeCases(tl);
        // FIXME: Allow some type combinations to unify, like different enums,
        // enums with strings, integers with doubles, maps with objects of
        // the correct type.
        if (!tsCases.isEmpty() && !tlCases.isEmpty() && !tsCases.equals(tlCases)) {
            return false;
        }
    }
    return true;
}
function isPartOfClique(c, clique) {
    for (var _i = 0, clique_1 = clique; _i < clique_1.length; _i++) {
        var cc = clique_1[_i];
        if (!canBeCombined(c, cc)) {
            return false;
        }
    }
    return true;
}
function combineClasses(graph, stringTypeMapping) {
    var unprocessedClasses = graph.allNamedTypesSeparated().classes.toArray();
    var cliques = [];
    // FIXME: Don't build cliques one by one.  Instead have a list of
    // cliques-in-progress and iterate over all classes.  Add the class
    // to the first clique that it's part of.  If there's none, make it
    // into a new clique.
    while (unprocessedClasses.length > 0) {
        var classesLeft = [];
        var clique = [unprocessedClasses[0]];
        for (var i = 1; i < unprocessedClasses.length; i++) {
            var c = unprocessedClasses[i];
            if (isPartOfClique(c, clique)) {
                clique.push(c);
            }
            else {
                classesLeft.push(c);
            }
        }
        if (clique.length > 1) {
            cliques.push(clique);
        }
        unprocessedClasses = classesLeft;
    }
    function makeCliqueClass(clique, builder) {
        Support_1.assert(clique.size > 0, "Clique can't be empty");
        var inferredNames = immutable_1.OrderedSet();
        var givenNames = immutable_1.OrderedSet();
        clique.forEach(function (c) {
            if (c.areNamesInferred) {
                inferredNames = inferredNames.union(c.names);
            }
            else {
                givenNames = givenNames.union(c.names);
            }
        });
        var areNamesInferred = givenNames.isEmpty();
        var properties = getCliqueProperties(clique, builder);
        return builder.getClassType(areNamesInferred ? inferredNames : givenNames, areNamesInferred, properties);
    }
    function getCliqueProperties(clique, builder) {
        var properties = immutable_1.OrderedMap();
        clique.forEach(function (c) {
            c.properties.forEach(function (t, name) {
                var p = properties.get(name);
                if (p) {
                    p[1] += 1;
                    // If one of the clique class's properties is nullable,
                    // the combined property must be nullable, too, so we
                    // just set it to this one.  Of course it can't be one of
                    // the properties that is just null.
                    if (t.kind !== "null") {
                        p[0] = t;
                    }
                    if (t.isNullable) {
                        p[2] = true;
                    }
                }
                else {
                    properties = properties.set(name, [t, 1, t.isNullable]);
                }
            });
        });
        return properties.map(function (_a, name) {
            var t = _a[0], count = _a[1], haveNullable = _a[2];
            var resultType = builder.reconstituteType(t);
            if (haveNullable || count < clique.size) {
                if (t.isNamedType()) {
                    resultType = builder.makeNullable(resultType, t.names, t.areNamesInferred);
                }
                else {
                    resultType = builder.makeNullable(resultType, name, true);
                }
            }
            return resultType;
        });
    }
    return graph.rewrite(stringTypeMapping, cliques, makeCliqueClass);
}
exports.combineClasses = combineClasses;
