"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var Type_1 = require("./Type");
var Support_1 = require("./Support");
var mapSizeThreshold = 20;
function shouldBeMap(properties) {
    // Only classes with a certain number of properties are inferred
    // as maps.
    if (properties.size < mapSizeThreshold) {
        return undefined;
    }
    // We need to handle three cases for maps (and the fourth case
    // where we leave the class as is):
    //
    // 1. All property types are null.
    // 2. Some property types are null or nullable.
    // 3. No property types are null or nullable.
    var nonNullCases = undefined;
    var isNullable = false;
    var canBeMap = true;
    // Check that all the property types are the same, modulo nullability.
    properties.forEach(function (t) {
        // The set of types first property can be, minus null.
        var nn = Type_1.nonNullTypeCases(t);
        if (!nn.isEmpty()) {
            if (nonNullCases !== undefined) {
                // The set of non-null cases for all other properties must
                // be the the same, otherwise we won't infer a map.
                if (!nn.toSet().equals(nonNullCases.toSet())) {
                    canBeMap = false;
                    return false;
                }
            }
            else {
                nonNullCases = nn;
            }
        }
        isNullable = isNullable || t.isNullable;
    });
    if (!canBeMap) {
        return undefined;
    }
    if (nonNullCases === undefined) {
        Support_1.assert(isNullable, "Non-nullable map candidate with no types");
        return [immutable_1.OrderedSet(), true];
    }
    return [nonNullCases, isNullable];
}
exports.shouldBeMap = shouldBeMap;
function replaceClass(setOfOneClass, builder) {
    var c = Support_1.defined(setOfOneClass.first());
    var properties = c.properties;
    var shouldBe = shouldBeMap(properties);
    if (shouldBe === undefined) {
        return Support_1.panic("We shouldn't be replacing class " + c.combinedName + " with a map");
    }
    var nonNulls = shouldBe[0], isNullable = shouldBe[1];
    // Now reconstitute all the types in the new graph.  TypeGraphs are
    // immutable, so any change in the graph actually means building a new
    // graph, and the types in the new graph are different objects.
    // Reconstituting a type means generating the "same" type in the new
    // type graph.  Except we don't get Type objects but TypeRef objects,
    // which is a type-to-be.
    var trefs = nonNulls.map(builder.reconstituteType);
    if (isNullable) {
        trefs = trefs.add(builder.getPrimitiveType("null"));
    }
    Support_1.assert(!trefs.isEmpty(), "We must have at least one type for the map properties");
    var valuesType;
    if (trefs.size === 1) {
        valuesType = Support_1.defined(trefs.first());
    }
    else {
        valuesType = builder.getUnionType(c.names, c.areNamesInferred, trefs);
    }
    return builder.getMapType(valuesType);
}
exports.replaceClass = replaceClass;
function inferMaps(graph, stringTypeMapping) {
    var allClasses = graph.allNamedTypesSeparated().classes;
    var classesToReplace = allClasses.filter(function (c) { return shouldBeMap(c.properties) !== undefined; }).toArray();
    return graph.rewrite(stringTypeMapping, classesToReplace.map(function (c) { return [c]; }), replaceClass);
}
exports.inferMaps = inferMaps;
