import { OrderedSet, OrderedMap, Collection } from "immutable";
import { TypeRef, TypeBuilder } from "./TypeBuilder";
export declare type PrimitiveStringTypeKind = "string" | "date" | "time" | "date-time";
export declare type PrimitiveTypeKind = "any" | "null" | "bool" | "integer" | "double" | PrimitiveStringTypeKind;
export declare type NamedTypeKind = "class" | "enum" | "union";
export declare type TypeKind = PrimitiveTypeKind | NamedTypeKind | "array" | "map";
export declare abstract class Type {
    readonly typeRef: TypeRef;
    readonly kind: TypeKind;
    constructor(typeRef: TypeRef, kind: TypeKind);
    isNamedType(): this is NamedType;
    readonly isStringType: boolean;
    readonly abstract children: OrderedSet<Type>;
    directlyReachableTypes<T>(setForType: (t: Type) => OrderedSet<T> | null): OrderedSet<T>;
    readonly abstract isNullable: boolean;
    abstract map(builder: TypeBuilder, f: (tref: TypeRef) => TypeRef): TypeRef;
    equals(other: any): boolean;
    hashCode(): number;
}
export declare class PrimitiveType extends Type {
    readonly kind: PrimitiveTypeKind;
    constructor(typeRef: TypeRef, kind: PrimitiveTypeKind);
    readonly children: OrderedSet<Type>;
    readonly isNullable: boolean;
    readonly isStringType: boolean;
    map(builder: TypeBuilder, _: (tref: TypeRef) => TypeRef): TypeRef;
}
export declare class ArrayType extends Type {
    private _itemsRef;
    readonly kind: "array";
    constructor(typeRef: TypeRef, _itemsRef?: TypeRef | undefined);
    setItems(itemsRef: TypeRef): void;
    private getItemsRef();
    readonly items: Type;
    readonly children: OrderedSet<Type>;
    readonly isNullable: boolean;
    map(builder: TypeBuilder, f: (tref: TypeRef) => TypeRef): TypeRef;
}
export declare class MapType extends Type {
    private _valuesRef;
    readonly kind: "map";
    constructor(typeRef: TypeRef, _valuesRef?: TypeRef | undefined);
    setValues(valuesRef: TypeRef): void;
    private getValuesRef();
    readonly values: Type;
    readonly children: OrderedSet<Type>;
    readonly isNullable: boolean;
    map(builder: TypeBuilder, f: (tref: TypeRef) => TypeRef): TypeRef;
}
export declare type NamesWithAlternatives = {
    names: OrderedSet<string>;
    alternatives: OrderedSet<string>;
};
export declare type NameOrNames = string | OrderedSet<string> | NamesWithAlternatives;
export declare abstract class NamedType extends Type {
    private _names;
    private _areNamesInferred;
    private _alternativeNames;
    constructor(typeRef: TypeRef, kind: NamedTypeKind, nameOrNames: NameOrNames, areNamesInferred: boolean);
    isNamedType(): this is NamedType;
    readonly names: OrderedSet<string>;
    readonly areNamesInferred: boolean;
    readonly alternativeNames: OrderedSet<string>;
    addNames(nameOrNames: NameOrNames, isInferred: boolean): void;
    clearInferredNames(): void;
    readonly combinedName: string;
    readonly proposedNames: OrderedSet<string>;
}
export declare class ClassType extends NamedType {
    private _propertyRefs;
    kind: "class";
    constructor(typeRef: TypeRef, names: NameOrNames, areNamesInferred: boolean, _propertyRefs?: OrderedMap<string, TypeRef> | undefined);
    setProperties(propertyRefs: OrderedMap<string, TypeRef>): void;
    private getPropertyRefs();
    readonly properties: OrderedMap<string, Type>;
    readonly sortedProperties: OrderedMap<string, Type>;
    readonly children: OrderedSet<Type>;
    readonly isNullable: boolean;
    map(builder: TypeBuilder, f: (tref: TypeRef) => TypeRef): TypeRef;
}
export declare class EnumType extends NamedType {
    readonly cases: OrderedSet<string>;
    kind: "enum";
    constructor(typeRef: TypeRef, names: NameOrNames, areNamesInferred: boolean, cases: OrderedSet<string>);
    readonly children: OrderedSet<Type>;
    readonly isNullable: boolean;
    readonly isStringType: boolean;
    map(builder: TypeBuilder, _: (tref: TypeRef) => TypeRef): TypeRef;
}
export declare class UnionType extends NamedType {
    private _memberRefs;
    kind: "union";
    constructor(typeRef: TypeRef, names: NameOrNames, areNamesInferred: boolean, _memberRefs?: OrderedSet<TypeRef> | undefined);
    setMembers(memberRefs: OrderedSet<TypeRef>): void;
    private getMemberRefs();
    readonly members: OrderedSet<Type>;
    readonly stringTypeMembers: OrderedSet<Type>;
    findMember: (kind: TypeKind) => Type | undefined;
    readonly children: OrderedSet<Type>;
    readonly isNullable: boolean;
    map(builder: TypeBuilder, f: (tref: TypeRef) => TypeRef): TypeRef;
    readonly sortedMembers: OrderedSet<Type>;
}
export declare function removeNullFromUnion(t: UnionType): [PrimitiveType | null, OrderedSet<Type>];
export declare function nullableFromUnion(t: UnionType): Type | null;
export declare function nonNullTypeCases(t: Type): OrderedSet<Type>;
export declare type SeparatedNamedTypes = {
    classes: OrderedSet<ClassType>;
    enums: OrderedSet<EnumType>;
    unions: OrderedSet<UnionType>;
};
export declare function separateNamedTypes(types: Collection<any, NamedType>): SeparatedNamedTypes;
export declare type StringTypeMatchers<U> = {
    dateType?: (dateType: PrimitiveType) => U;
    timeType?: (timeType: PrimitiveType) => U;
    dateTimeType?: (dateTimeType: PrimitiveType) => U;
};
export declare function matchTypeExhaustive<U>(t: Type, anyType: (anyType: PrimitiveType) => U, nullType: (nullType: PrimitiveType) => U, boolType: (boolType: PrimitiveType) => U, integerType: (integerType: PrimitiveType) => U, doubleType: (doubleType: PrimitiveType) => U, stringType: (stringType: PrimitiveType) => U, arrayType: (arrayType: ArrayType) => U, classType: (classType: ClassType) => U, mapType: (mapType: MapType) => U, enumType: (enumType: EnumType) => U, unionType: (unionType: UnionType) => U, dateType: (dateType: PrimitiveType) => U, timeType: (timeType: PrimitiveType) => U, dateTimeType: (dateTimeType: PrimitiveType) => U): U;
export declare function matchType<U>(t: Type, anyType: (anyType: PrimitiveType) => U, nullType: (nullType: PrimitiveType) => U, boolType: (boolType: PrimitiveType) => U, integerType: (integerType: PrimitiveType) => U, doubleType: (doubleType: PrimitiveType) => U, stringType: (stringType: PrimitiveType) => U, arrayType: (arrayType: ArrayType) => U, classType: (classType: ClassType) => U, mapType: (mapType: MapType) => U, enumType: (enumType: EnumType) => U, unionType: (unionType: UnionType) => U, stringTypeMatchers?: StringTypeMatchers<U>): U;
export declare function matchCompoundType(t: Type, arrayType: (arrayType: ArrayType) => void, classType: (classType: ClassType) => void, mapType: (mapType: MapType) => void, unionType: (unionType: UnionType) => void): void;
